#pragma once

#include <list>
#include <map>
#include <memory>
#include <string>
#include <tuple>
#include <utility>
#include <vector>

#include <dune/common/fvector.hh>
#include <dune/common/fmatrix.hh>
#include <dune/grid/common/grid.hh>

#include <amdis/AdaptInfo.hpp>
#include <amdis/CreatorInterface.hpp>
#include <amdis/CreatorMap.hpp>
#include <amdis/DirichletBC.hpp>
#include <amdis/Flag.hpp>
#include <amdis/Initfile.hpp>
#include <amdis/LinearAlgebra.hpp>
#include <amdis/LinearSolvers.hpp>
#include <amdis/LocalAssemblerList.hpp>
#include <amdis/Marker.hpp>
#include <amdis/Mesh.hpp>
#include <amdis/ProblemStatBase.hpp>
#include <amdis/ProblemStatTraits.hpp>
#include <amdis/StandardProblemIteration.hpp>

#include <amdis/common/TupleUtility.hpp>
#include <amdis/common/Utility.hpp>

#include <amdis/GridFunctions.hpp>
#include <amdis/gridfunctions/DOFVectorView.hpp>

#include <amdis/io/FileWriterInterface.hpp>

#include <amdis/utility/TreeData.hpp>
#include <amdis/utility/TreePath.hpp>

namespace AMDiS
{
  // forward declaration
  template <class Traits>
  class ProblemInstat;

  template <class Traits>
  class ProblemStat
      : public ProblemStatBase
      , public StandardProblemIteration
  {
    using Self = ProblemStat;

    friend class ProblemInstat<Traits>;

  public: // typedefs and static constants

    using GlobalBasis = typename Traits::GlobalBasis;
    using GridView = typename GlobalBasis::GridView;
    using Grid     = typename GridView::Grid;
    using Element  = typename GridView::template Codim<0>::Entity;
    using WorldVector = typename Element::Geometry::GlobalCoordinate;

    /// Dimension of the grid
    static constexpr int dim = Grid::dimension;

    /// Dimension of the world
    static constexpr int dow = Grid::dimensionworld;

    using SystemMatrix = DOFMatrix<GlobalBasis, GlobalBasis, double>;
    using SystemVector = DOFVector<GlobalBasis, double>;

    using LinearSolverType = LinearSolverInterface<typename SystemMatrix::BaseMatrix, typename SystemVector::BaseVector>;

  public:
    /**
     * \brief Constructor. Takes the name of the problem that is used to
     * access values corresponding to this problem in the parameter file.
     **/
    explicit ProblemStat(std::string name)
      : StandardProblemIteration(dynamic_cast<ProblemStatBase&>(*this))
      , name_(std::move(name))
    {}

    /// Constructor taking additionally a reference to a grid that is used
    /// instead of the default created grid, \ref ProblemStat
    ProblemStat(std::string name, Grid& grid)
      : ProblemStat(std::move(name))
    {
      grid_ = Dune::stackobject_to_shared_ptr(grid);
      Parameters::get(name_ + "->mesh", gridName_);
    }

    /// \brief Constructor taking a grid reference and a basis reference.
    /// Stores pointers to both.
    ProblemStat(std::string name, Grid& grid, GlobalBasis& globalBasis)
      : ProblemStat(std::move(name), grid)
    {
      globalBasis_ = Dune::stackobject_to_shared_ptr(globalBasis);
      initGlobalBasis(*globalBasis_);
    }


    /**
     * \brief Initialisation of the problem.
     *
     * Parameters read in initialize() for problem with name 'PROB'
     *   MESH[0]->global refinements:  nr of initial global refinements
     **/
    void initialize(Flag initFlag,
                    Self* adoptProblem = nullptr,
                    Flag adoptFlag = INIT_NOTHING);


    /// Adds an operator to \ref A.
    /** @{ */
    template <class Operator, class RowTreePath = RootTreePath, class ColTreePath = RootTreePath>
    void addMatrixOperator(Operator const& op, RowTreePath row = {}, ColTreePath col = {})
    {
      systemMatrix_->addOperator(tag::element_operator<Element>{}, op, row, col);
    }

    // operator evaluated on the boundary
    template <class Operator, class RowTreePath = RootTreePath, class ColTreePath = RootTreePath>
    void addMatrixOperator(BoundaryType b, Operator const& op, RowTreePath row = {}, ColTreePath col = {})
    {
      systemMatrix_->addOperator(tag::boundary_operator<typename GridView::Intersection>{b}, op, row, col);
    }
    /** @} */


    /// Adds an operator to \ref rhs.
    /** @{ */
    template <class Operator, class TreePath = RootTreePath>
    void addVectorOperator(Operator const& op, TreePath path = {})
    {
      rhs_->addOperator(tag::element_operator<Element>{}, op, path);
    }

    // operator evaluated on the boundary
    template <class Operator, class TreePath = RootTreePath>
    void addVectorOperator(BoundaryType b, Operator const& op, TreePath path = {})
    {
      rhs_->addOperator(tag::boundary_operator<typename GridView::Intersection>{b}, op, path);
    }
    /** @} */


    /// Adds a Dirichlet boundary condition
    template <class Predicate, class RowTreePath, class ColTreePath, class Values>
    void addDirichletBC(Predicate const& predicate,
                        RowTreePath row, ColTreePath col,
                        Values const& values);

  public:

    /// Implementation of \ref ProblemStatBase::solve
    virtual void solve(AdaptInfo& adaptInfo,
                       bool createMatrixData = true,
                       bool storeMatrixData = false) override;

    /// Implementation of \ref ProblemStatBase::buildAfterCoarse
    virtual void buildAfterAdapt(AdaptInfo& adaptInfo,
                                 Flag flag,
                                 bool asmMatrix = true,
                                 bool asmVector = true) override;

    /// Writes output files.
    void writeFiles(AdaptInfo& adaptInfo, bool force = false);


  public: // get-methods

    /// Returns a reference to system-matrix, \ref systemMatrix_
    SystemMatrix&       getSystemMatrix()       { return *systemMatrix_; }
    SystemMatrix const& getSystemMatrix() const { return *systemMatrix_; }

    /// Returns a reference to the solution vector, \ref solution_
    SystemVector&       getSolutionVector()       { return *solution_; }
    SystemVector const& getSolutionVector() const { return *solution_; }

    /// Return a reference to the rhs system-vector, \ref rhs
    SystemVector&       getRhsVector()       { return *rhs_; }
    SystemVector const& getRhsVector() const { return *rhs_; }


    /// Return a mutable view to a solution component
    template <class TreePath = RootTreePath>
    auto getSolution(TreePath const& path = {})
    {
      auto&& tp = makeTreePath(path);
      return makeDOFVectorView(*solution_, tp);
    }

    /// Return a const view to a solution component
    template <class TreePath = RootTreePath>
    auto getSolution(TreePath const& path = {}) const
    {
      auto&& tp = makeTreePath(path);
      return makeDOFVectorView(*solution_, tp);
    }


    /// Return a reference to the linear solver, \ref linearSolver
    LinearSolverType&       getSolver()       { return *linearSolver_; }
    LinearSolverType const& getSolver() const { return *linearSolver_; }

    /// Set a new linear solver for the problem
    void setSolver(std::shared_ptr<LinearSolverType> const& solver)
    {
      linearSolver_ = solver;
    }

    /// Return a reference to the grid, \ref grid
    Grid&       grid()       { return *grid_; }
    Grid const& grid() const { return *grid_; }

    /// Set the grid. Stores pointer and initializes feSpaces
    /// matrices and vectors, as well as the file-writer.
    void setGrid(std::shared_ptr<Grid> const& grid)
    {
      grid_ = grid;

      createGlobalBasis();
      createMatricesAndVectors();
      createFileWriter();
    }

    /// Return the gridView of the leaf-level
    GridView const& gridView() { return globalBasis_->gridView(); }

    /// Return the \ref globalBasis_
    GlobalBasis const& globalBasis() const { return *globalBasis_; }

    /// Return the \ref globalBasis_
    GlobalBasis& globalBasis() { return *globalBasis_; }


    /// Implementation of \ref ProblemStatBase::getName
    virtual std::string getName() const override
    {
      return name_;
    }

  protected: // initialization methods

    template <class T, class GV>
    using HasCreate = decltype(T::create(std::declval<GV>()));

    void createGlobalBasis()
    {
      createGlobalBasis(Dune::Std::is_detected<HasCreate,Traits,GridView>{});
      initGlobalBasis(*globalBasis_);
    }

    void createGlobalBasis(std::true_type)
    {
      assert( bool(grid_) );
      static_assert(std::is_same<GridView, typename Grid::LeafGridView>::value, "");
      globalBasis_ = std::make_shared<GlobalBasis>(Traits::create(grid_->leafGridView()));
    }

    void createGlobalBasis(std::false_type)
    {
      error_exit("Cannot create GlobalBasis from type. Pass a BasisCreator instead!");
    }

    void createGrid()
    {
      gridName_ = "";
      Parameters::get(name_ + "->mesh", gridName_);
      test_exit(!gridName_.empty(), "No grid name specified for '{}->mesh'!", name_);

      grid_ = MeshCreator<Grid>::create(gridName_);

      msg("Create grid:");
      msg("#elements = {}"   , grid_->size(0));
      msg("#faces/edges = {}", grid_->size(1));
      msg("#vertices = {}"   , grid_->size(dim));
      msg("");
    }

    void initGlobalBasis(GlobalBasis const& globalBasis)
    {
      constraints_.init(*globalBasis_, *globalBasis_);
    }

    void createMatricesAndVectors()
    {
      systemMatrix_ = std::make_shared<SystemMatrix>(*globalBasis_, *globalBasis_);
      solution_ = std::make_shared<SystemVector>(*globalBasis_);
      rhs_ = std::make_shared<SystemVector>(*globalBasis_);

      auto localView = globalBasis_->localView();
      AMDiS::forEachNode_(localView.tree(), [&,this](auto const& node, auto treePath)
      {
        std::string i = to_string(treePath);
        estimates_[i].resize(globalBasis_->gridView().indexSet().size(0));
        for (std::size_t j = 0; j < estimates_[i].size(); j++)
          estimates_[i][j] = 0.0; // TODO: Remove when estimate() is implemented
      });
    }

    void createSolver()
    {
      std::string solverName = "cg";
      Parameters::get(name_ + "->solver->name", solverName);

      auto solverCreator
        = named(CreatorMap<LinearSolverType>::getCreator(solverName, name_ + "->solver->name"));

      linearSolver_ = solverCreator->create(name_ + "->solver");
    }

    void createMarker();

    void createFileWriter();


  public: // implementation of iteration interface methods

    /**
      * \brief Determines the execution order of the single adaption steps.
      *
      * If adapt is true, mesh adaption will be performed. This allows to avoid
      * mesh adaption, e.g. in timestep adaption loops of timestep adaptive
      * strategies.
      *
      * Implementation of \ref StandardProblemIteration::oneIteration.
      **/
    virtual Flag oneIteration(AdaptInfo& adaptInfo, Flag toDo = FULL_ITERATION) override
    {
      return StandardProblemIteration::oneIteration(adaptInfo, toDo);
    }

    /// Implementation of \ref ProblemStatBase::estimate.
    virtual void estimate(AdaptInfo& adaptInfo) override { /* Does nothing. */ }

    /// Implementation of \ref ProblemStatBase::refineMesh.
    virtual Flag adaptGrid(AdaptInfo& adaptInfo) override;

    /// Implementation of \ref ProblemStatBase::markElements.
    virtual Flag markElements(AdaptInfo& adaptInfo) override;


  private:
    /// Name of this problem.
    std::string name_;

    /// Grid of this problem.
    std::shared_ptr<Grid> grid_; // TODO: generalize to multi-mesh problems

    /// Number of grids
    int nGrids = 1;

    /// Name of the grid
    std::string gridName_ = "none";

    /// FE spaces of this problem.
    std::shared_ptr<GlobalBasis> globalBasis_;

    /// A FileWriter object
    std::list<std::shared_ptr<FileWriterInterface>> filewriter_;

    /// Pointer to the adaptation markers
    std::list<std::shared_ptr<Marker<Traits> > > marker_;

    /// An object of the linearSolver Interface
    std::shared_ptr<LinearSolverType> linearSolver_;

    /// A block-matrix that is filled during assembling
    std::shared_ptr<SystemMatrix> systemMatrix_;

    /// A block-vector with the solution components
    std::shared_ptr<SystemVector> solution_;

    /// A vector with the local element error estimates
    /// for each node in the basis tree, indexed by [to_string(treePath)][element index]
    std::map<std::string, std::vector<double> > estimates_;

    /// A block-vector (load-vector) corresponding to the right.hand side
    /// of the equation, filled during assembling
    std::shared_ptr<SystemVector> rhs_;


  private: // some internal data-structures

    Constraints<GlobalBasis, GlobalBasis> constraints_;
  };

#if DUNE_HAVE_CXX_CLASS_TEMPLATE_ARGUMENT_DEDUCTION
  // Deduction rule
  template <class Grid, class GlobalBasis>
  ProblemStat(std::string name, Grid& grid, GlobalBasis& globalBasis)
    -> ProblemStat<DefaultProblemTraits<GlobalBasis>>;
#endif

  // Generator for ProblemStat with given Grid and GlobalBasis
  template <class Grid, class GlobalBasis>
  ProblemStat<DefaultProblemTraits<GlobalBasis>>
  makeProblemStat(std::string name, Grid& grid, GlobalBasis& globalBasis)
  {
    return {std::move(name), grid, globalBasis};
  }


#ifndef AMDIS_NO_EXTERN_PROBLEMSTAT
  // extern template class ProblemStat<YaspGridBasis<2,1>>;
  // extern template class ProblemStat<YaspGridBasis<2,2>>;
  // extern template class ProblemStat<YaspGridBasis<2,1,2>>;
#endif

} // end namespace AMDiS

#include "ProblemStat.inc.hpp"

