#pragma once

#include <tuple>
#include <type_traits>

#include <dune/common/std/apply.hh>

#include <amdis/Operations.hpp>
#include <amdis/common/IndexSeq.hpp>
#include <amdis/common/Loops.hpp>
#include <amdis/common/Mpl.hpp>
#include <amdis/gridfunctions/GridFunctionConcepts.hpp>

namespace AMDiS
{
  namespace Impl
  {
    template <class T0, class... Ts>
    struct DomainType
    {
      using type = typename T0::Domain;
      // static_assert( all_of_v< std::is_same<type, typename DomainType<Ts>::type>::value... >,
      //   "All GridFunctions must have the same Domain." );
    };

    template <class T0, class... Ts>
    struct EntitySetType
    {
      using type = typename T0::EntitySet;
      // static_assert( all_of_v< std::is_same<type, typename EntitySetType<Ts>::type>::value... >,
      //   "All GridFunctions must have the same EntitySet." );
    };

  } // end namespace Impl


  // forward declaration
  template <class Signatur, class Functor, class... LocalFunctions>
  class FunctorLocalFunction;

  // implementation
  template <class R, class D, class Functor, class... LocalFunctions>
  class FunctorLocalFunction<R(D), Functor, LocalFunctions...>
  {
  public:
    using Range = R;
    using Domain = D;

    enum { hasDerivative = true };

    template <class LocalFct>
    struct LocalFunctionWrapper
    {
      template <class GridFct>
      LocalFunctionWrapper(GridFct const& gridFct)
        : localFct_{localFunction(gridFct)}
      {}

      LocalFct const& operator*() const { return localFct_; }
      LocalFct      & operator*()       { return localFct_; }

      LocalFct localFct_;
    };

  public:
    /// Constructor. Stores copies of the functor and localFunction(gridfunction)s.
    template <class... GridFcts>
    FunctorLocalFunction(Functor const& fct, GridFcts&&... gridFcts)
      : fct_{fct}
      , localFcts_{std::forward<GridFcts>(gridFcts)...}
    {}

    /// Calls \ref bind for all localFunctions
    template <class Element>
    void bind(Element const& element)
    {
      forEach(localFcts_, [&](auto& localFct) {
        (*localFct).bind(element);
      });
    }

    /// Calls \ref unbind for all localFunctions
    void unbind()
    {
      forEach(localFcts_, [&](auto& localFct) {
        (*localFct).unbind();
      });
    }

    /// Applies the functor \ref fct_ to the evaluated localFunctions
    Range operator()(Domain const& x) const
    {
      using Dune::Std::apply;
      return apply([&](auto&&... localFct) { return fct_((*localFct)(x)...); }, localFcts_);
    }

  public:
    /// Return the stored functor
    Functor const& fct() const
    {
      return fct_;
    }

    auto const& localFcts() const
    {
      return localFcts_;
    }

  private:
    Functor fct_;
    std::tuple<LocalFunctionWrapper<LocalFunctions>...> localFcts_;
  };

  // Derivative of the LocalFunction of a FunctorGridFunction, utilizing
  // the chain-rule. Only available of the functor provides partial derivatives.
  /*
   * \f$ d_x(f(lf(x)...)) = \sum_i d_i(f)[lf(x)...] * derivative(lf[i]) \f$
   */
  template <class Sig, class F, class... LFs,
    REQUIRES(Concepts::HasPartial<F>)>
  auto derivative(FunctorLocalFunction<Sig,F,LFs...> const& lf)
  {
    auto index_seq = std::index_sequence_for<LFs...>{};

    // d_i(f)[lgfs...] * lgfs_i
    auto term_i = [&](auto const _i)
    {
      using Dune::Std::apply;
      auto di_f = apply([&](auto const&... lgfs) {
        return makeFunctorGridFunction(partial(lf.fct(), _i), (*lgfs)...);
      }, lf.localFcts());

      using std::get;
      auto const& lgfs_i = *get<decltype(_i)::value>(lf.localFcts());
      return makeFunctorGridFunction(Operation::Multiplies{}, di_f, derivative(lgfs_i));
    };

    // sum_i [ d_i(f)[lgfs...] * derivative(lgfs_i)
    using Dune::Std::apply;
    auto gridFct = apply([&](auto const... _i)
    {
      return makeFunctorGridFunction(Operation::Plus{}, term_i(_i)...);
    }, index_seq);

    return localFunction(gridFct);
  }


  // Calculate the polynomial order for functors f providing a free functions
  // order(f, degs...), where degs are the order of the LocalFunctions.
  template <class Sig, class F, class... LFs,
    REQUIRES(Concepts::HasFunctorOrder<F,sizeof...(LFs)>
             && all_of_v<Concepts::HasOrder<LFs>...>)>
  int order(FunctorLocalFunction<Sig,F,LFs...> const& lf)
  {
    using Dune::Std::apply;
    return apply([&lf](auto const&... lgfs) { return order(lf.fct(), order(*lgfs)...); },
      lf.localFcts());
  }


  /**
    * \addtogroup GridFunctions
    * @{
    **/

  /// \brief A Gridfunction that applies a functor to the evaluated Gridfunctions
  /**
   * Composition of GridFunctions `g_i` by applying a functor `f` locally, i.e. locally it is evaluated
   * \f$ f(g_0(x), g_1(x), ...) \f$
   *
   * \tparam Functor           The type of the outer functor `f`
   * \tparam GridFunctions...  The GridFunction types of `g_i`
   *
   * Requirements:
   * - `arity(f) == sizeof...(GridFunctions)`
   * - `arity(g_i) == arity(g_j) for i != j`
   * - `g_i` models concept \ref GridFunction
   **/
  template <class Functor, class... GridFunctions>
  class FunctorGridFunction
  {
  public:
    /// The result type of the functor when applied to the grid-functions
    using Range = decltype(std::declval<Functor>()(std::declval<typename GridFunctions::Range>()...));
    /// The argument type that can be applied to the grid-functions
    using Domain = typename Impl::DomainType<GridFunctions...>::type;
    /// The set of entities this grid-function binds to
    using EntitySet = typename Impl::EntitySetType<GridFunctions...>::type;

    enum { hasDerivative = false };

  private:
    template <class GridFct>
    using LocalFct = std::decay_t<decltype(localFunction(std::declval<GridFct>()))>;

    using RawRange = std::decay_t<Range>;
    using LocalDomain = typename EntitySet::LocalCoordinate;

  public:
    using LocalFunction = FunctorLocalFunction<RawRange(LocalDomain), Functor, LocalFct<GridFunctions>...>;

    /// Constructor. Stores copies of the functor and gridfunctions.
    template <class... GridFcts>
    explicit FunctorGridFunction(Functor const& fct, GridFcts&&... gridFcts)
      : fct_{fct}
      , gridFcts_{std::forward<GridFcts>(gridFcts)...}
    {}

    /// Applies the functor to the evaluated gridfunctions
    Range operator()(Domain const& x) const
    {
      using Dune::Std::apply;
      return apply([&](auto&&... gridFct) { return fct_(gridFct(x)...); }, gridFcts_);
    }

    /// Return the stored \ref EntitySet of the first GridFunction
    EntitySet const& entitySet() const
    {
      using std::get;
      return get<0>(gridFcts_).entitySet();
    }

    LocalFunction localFunction() const
    {
      using Dune::Std::apply;
      return apply([&](auto const&... gridFcts) { return LocalFunction{fct_, gridFcts...}; }, gridFcts_);
    }

  private:
    Functor fct_;
    std::tuple<GridFunctions...> gridFcts_;
  };


  /// \brief Creates a LocalFunction from the LocalFunctions of the GridFunctions. \relates FunctorLocalFunction
  template <class F, class... GFs>
  auto localFunction(FunctorGridFunction<F,GFs...> const& gf)
  {
    return gf.localFunction();
  }


  // Generator function for FunctorGridFunction expressions
  template <class Functor, class... GridFcts>
  auto makeFunctorGridFunction(Functor const& f, GridFcts const&... gridFcts)
  {
    static_assert(all_of_v<Concepts::GridFunction<GridFcts>...>,
      "All passed parameters must be GridFunctions.");
    static_assert(Concepts::Callable<Functor, typename GridFcts::Range...>,
      "Range types of grid functions are not compatible with the functor.");
    return FunctorGridFunction<Functor, GridFcts...>{f, gridFcts...};
  }

  // PreGridFunction related to FunctorGridFunction.
  template <class Functor, class... PreGridFunctions>
  struct FunctorPreGridFunction
  {
    using Self = FunctorPreGridFunction;

    struct Creator
    {
      template <class GridView>
      static auto create(Self const& self, GridView const& gridView)
      {
        return Dune::Std::apply([&](auto const&... pgf) {
          return makeFunctorGridFunction(self.fct_,
            makeGridFunction(pgf, gridView)...);
        }, self.preGridFcts_);
      }
    };

    template <class... PreGridFcts>
    explicit FunctorPreGridFunction(Functor const& fct, PreGridFcts&&... pgfs)
      : fct_{fct}
      , preGridFcts_{std::forward<PreGridFcts>(pgfs)...}
    {}

  private:
    Functor fct_;
    std::tuple<PreGridFunctions...> preGridFcts_;
  };

  namespace Traits
  {
    template <class Functor, class... PreGridFcts>
    struct IsPreGridFunction<FunctorPreGridFunction<Functor, PreGridFcts...>>
      : std::true_type {};
  }


  /// \brief Generator function for FunctorGridFunction. \relates FunctorGridFunction
  /**
   * Applies the functor `f` to the grid-functions `gridFcts...`. See \ref FunctorGridFunction.
   *
   * **Examples:**
   * - `invokeAtQP([](Dune::FieldVector<double, 2> const& x) { return two_norm(x); }, X());`
   * - `invokeAtQP([](double u, auto const& x) { return u + x[0]; }, 1.0, X());`
   * - `invokeAtQP(Operation::Plus{}, X(0), X(1));`
   **/
  template <class Functor, class... PreGridFcts>
  auto invokeAtQP(Functor const& f, PreGridFcts&&... gridFcts)
  {
    return FunctorPreGridFunction<Functor, std::decay_t<PreGridFcts>...>{f, std::forward<PreGridFcts>(gridFcts)...};
  }

  /** @} **/

} // end namespace AMDiS
